module Constants where

import Graphics.Gloss
{-

This file defines the constants of the simulation and the display options
By all means edit these and find interesting combinations
Be aware that the simulation may be instable upon certain conditions, or simply will not function, for example:

Defining a collision radius greater than or equal to half of the anchor length, this causes the nodes to be
 unable to move.
Defining a collision radius too small for the anchor length, an appropriate collision radius is about 40% of the
 anchor length. A small collision radius causes the structure to collapse inside of itself. Note that this is not
 exactly an issue with the simulation, but an inherant limitation with the cross-hatch node and spring structure.
 (Using the standard node-node collision system that is)
A damping constant too low for the spring constant, causing an explotion in spring forces. This explosion is a
 result of the integration technique being used(1), whereas more stable integration techniques(2) require more
 computing power.
 (1) Integration techniques are used due to the limited number of frames per second, if we had an unlimited fps,
     and a delta time of zero, we would have no issues. 
     The current technique is semi-implicit euler, which is a fancy way of saying update velocity first then
      position. Euler refers to the Euler method of approximating a first order differential equation.
 (2) More advanced techniques such as Runge-Kutte 2 or 4 provide a better approximation, but are more expensive.
     Verlet integration does however exist, which is more stable without a difference in computation, but for
     this project it is not exactly needed.
Constants very high or very small, these values can be exacerbated by numerical instability and may break the simulation.
-}

structureWidth :: Int
structureWidth = 14

structureHeight :: Int
structureHeight = 14

massC :: Float 
massC = 2.0

--This is the scalar of the gravity vector applied to the velocity every frame
gravityC :: Float
gravityC = 1.2

radiusC :: Float
radiusC = 2.0

--This radius is seperate as it defines the radius at which the nodes should collide with eachover.
--This does not include boundary collision.
--They have been separated to give more design control over the nodes.
collisionRadiusC :: Float
collisionRadiusC = 35.0 * 0.4

--Damping coefficient
dampingC :: Float
dampingC = 5.0

--Spring constant used in Hooke's law
springConstantC :: Float
springConstantC = 35.0

anchorLengthC :: Float
anchorLengthC = 35.0

--Friction 0.0->1.0
frictionConstantC :: Float
frictionConstantC = 0.01

--The offset of the top left of the structure
--Note that the coordinates are based off of the top left of the boundaries, which is different from default gloss.
origin :: (Float, Float)
origin = (150.0, 100.0)

--Display options

--Sizes of the boundary
screenX :: Float
screenX = 900.0

screenY :: Float 
screenY = 900.0

showNodes :: Bool
showNodes = True

filledNodes :: Bool
filledNodes = True

--Default is Push (1.0 0.0 0.0), Pull (0.0 0.5 1.0)
pullColour :: (Float, Float, Float)
pullColour = (0.0, 0.5, 1.0) --Blue/Cyan

pushColour :: (Float, Float, Float)
pushColour = (1.0, 0.0, 0.0) --Red

nodeColour :: (Float, Float, Float)
nodeColour = (0.4, 0.4, 0.4)

backgroundColor :: (Float, Float, Float)
backgroundColor = (0.04, 0.025, 0.05)
